<?php

/**
 * @file
 * Mover export module Drush integration.
 */

/**
 * Implementation of hook_drush_command().
 *
 * @see drush_parse_command()
 * for a list of recognized keys.
 *
 * @return
 *   An associative array describing your command(s).
 */
function mover_export_drush_command() {
  $items = array();

  $items['mover-export'] = array(
    'callback' => 'mover_export_drush_export',
    'description' => "Export content from your site using the Mover module.",
    'drupal dependencies' => array('mover_export'),
    'options' => array(
      '--process' => "Specify the name of a process to run",
      '--progress' => "Specify whether or not show progress while running",
      '--start' => "Specify an item number to start with",
      '--number' => "Specify the number of items to export",
      '--continue' => "Continue an export process where the last one left off",
    ),
  );

  return $items;
}

/**
 * Implementation of hook_drush_help().
 */
function mover_export_drush_help($section) {
  switch ($section) {
    case 'drush:mover-export':
      return dt("Export content from your site using the Mover module.");
  }
}

function mover_export_drush_export() {

  // Check the main export directory
  $directory = variable_get('mover_export_directory', file_directory_path() .'/export');
  $directory = rtrim($directory, '/\\');
  if (!file_check_directory($directory, TRUE)) {
    drush_print(dt('The directory %directory does not exist.', array('%directory' => $directory)));
    return;
  }

  // Get input options
  $process = drush_get_option('process');
  $progress = drush_get_option('progress');
  $continue = drush_get_option('continue');
  $start = drush_get_option('start');
  $number = drush_get_option('number');

  // Get the current process
  $process_info = mover_export_get_process($process);
  if (!$process_info) {
    $process_names = array();
    $processes = module_invoke_all('mover_export_processes');
    foreach ($processes as $process_name => $info) {
      $process_names[] = $process_name;
    }
    drush_print(dt('You must specify a process to run using the --process=x option with one of these process names: @processes', array(
      '@processes' => implode(', ', $process_names)
    )));
    return;
  }

  // Figure out what process number to start with
  if ($start && $continue) {
    drush_print(dt('You cannot use the --start and --continue options at the same time.'));
    return;
  }
  if ($continue) {
    $start = variable_get('mover_export_start', 1);
    if ($start == 1) {
      drush_print(dt('There is not a current export process to continue.'));
      return;
    }
  }
  elseif (!$start) {
    $start = 1;
  }

  // Get the number of items to process
  $number = ($number) ? $number : variable_get('mover_export_number', 1000);

  // Save the current variables for later
  variable_set('mover_export_number', $number);
  variable_set('mover_export_start', $start);

  // Display a message
  drush_print(dt('Processing up to @number items starting at @start. Use --start=x and --number=y options to modify. Use the --progress option to watch the progress.', array(
      '@number' => $number,
      '@start' => $start)));

  // Certain types of processes can be handled internally (by this module)
  // Other types of processes are handled externally by other modules
  $function = $process_info['operations callback'];
  if (function_exists($function)) {
    $operations = $function($process_info['process callback']);
  }
  else {
    drush_print(dt('Function %function not found!', array('%function' => $function)));
    return;
  }

  // Run all operations
  $context = array('drush' => TRUE);
  foreach ($operations as $i => $operation) {
    if ($progress) {
      drush_print(dt('Processing @num of @total (@file, memory: @memory)', array(
        '@num' => $start + $i,
        '@total' => variable_get('mover_export_total', count($operations)),
        '@file' => $operation[1][0],
        '@memory' => round(memory_get_usage() / 1048576) . ' MB'
      )));
    }
    $operation[1][] = &$context;
    call_user_func_array($operation[0], $operation[1]);
    _drush_log_drupal_messages();
  }

  if ((variable_get('mover_export_start', 1) > 1) &&
      (variable_get('mover_export_start', 1) < variable_get('mover_export_total', 0))) {
    drush_print(dt("There are still @num items left to process. Run this command again to continue.",
      array('@num' => variable_get('mover_export_total', 0) - variable_get('mover_export_start', 1) + 1)));
  }

  if (!$operations) {
    drush_print(dt("There isn't any content to process for this operation!"));
  }
}
