<?php

/**
 * @file
 * Handles importing of referenced data (e.g. data from node reference and user
 * reference CCK fields).
 */

/**
 * Builds an array of batch operations that loops through all unprocessed
 * reference maps and let's other functions and other modules process them.
 */
function mover_import_references_operations($process) {
  $operations = array();

  // Variables for this run
  $start = variable_get('mover_import_start', 1);
  $number = variable_get('mover_import_number', 1000);
  $i = 0;

  // Get information about the process
  $process_info = mover_import_get_process($process);

  // Re-Map references
  // This operation links up node references, user references, and maybe other types of custom references!
  // This couldn't be done earlier because the nodes/user/etc that are being references may not have been
  // imported before now.
  //
  // Start by getting the types of content that need references mapping (node, user, etc)
  $result = db_query("SELECT DISTINCT m.type FROM {mover_map_references} mr
    INNER JOIN {mover_map} m ON m.map_id=mr.map_id
    WHERE mr.complete=0;");
  while ($types = db_fetch_object($result)) {
    $type = $types->type; // node, user, term, etc.

    // Now get all the nodes, users, etc that have references which need mapping. We're
    // building a list of operations for NIDs, UIDs, etc that will need to be updated.
    $result2 = db_query("SELECT DISTINCT m.new_content_id FROM {mover_map_references} mr
      INNER JOIN {mover_map} m ON m.map_id=mr.map_id
      WHERE mr.complete=0 AND m.type='%s'
      ORDER BY m.new_content_id", $type);
    while ($content_ids = db_fetch_object($result2)) {
      $id = $content_ids->new_content_id;

      // Now build a batch operations queue
      $i++;
      if ($i <= $number) {
        $function = $process_info['process callback'];
        if (function_exists($function)) {
          $operations[] = array($function, array($type, $id));
        }
        else {
          drupal_set_message(t('Function %function not found!', array(
            '%function' => $function,
          )), 'error');
        }
      }
    }
  }

  // Save the number of items that should be / will be imported
  variable_set('mover_import_total', $i);

  return $operations;
}

/**
 * References processor for nodes (not just node references, but also user references and any other
 * type of reference that a node might have to another piece of content). This function is
 * called after all other importing is done, so that the referenced content IDs will be valid.
 *
 * @param $type
 *   The type of content which has references to be updated, e.g. node, or user
 * @param $id
 *   The id of the content which has references to be updated, e.g. a NID or UID
 */
function mover_import_references_process($type, $id, &$context) {

  // Load references to be mapped for this content
  $references = array();
  $result = db_query("SELECT mr.* FROM {mover_map_references} mr
    INNER JOIN {mover_map} m ON m.map_id=mr.map_id
    WHERE m.type='%s' AND m.new_content_id=%d
    ORDER BY reference_type, reference_id", $type, $id);
  while ($data = db_fetch_object($result)) {
    $map_id = $data->map_id;
    $references[$data->reference_type][$data->field][] = $data->old_reference_id;
  }

  // This shouldn't happen, because of how this function is called, but just to
  // be safe...
  if (!$map_id) {
    drupal_set_message(t('No map ID for reference mapping of @type @id', array(
      '@type' => $type,
      '@id' => $id,
    )), 'error');
    mover_import_increment_counter($context);
    return;
  }

  // Let other modules (including this module) process reference mapping
  module_invoke_all('mover_import_reference_mapping', $map_id, $type, $id, $references);

  //drupal_set_message(t('Updated references for @type @id', array('@type' => $type, '@id' => $id)));
  $context['message'] = t('Updated references for @type @id', array('@type' => $type, '@id' => $id));
  mover_import_increment_counter($context);
}

/**
 * Implementation of hook_mover_import_reference_mapping().
 *
 * This module handles node reference and user reference mapping. Other
 * modules can implement this hook to handle other types of references.
 */
function mover_import_mover_import_reference_mapping($map_id, $type, $id, $references) {
  switch ($type) {
    case 'node':
      // Process node and user reference fields on nodes
      if (is_array($references['node']) || is_array($references['user'])) {
        $node = node_load($id, NULL, TRUE);
        node_invoke_nodeapi($node, 'import_load');
        node_invoke_nodeapi($node, 'import_references', $references);
        node_save($node);

        // Mark all these references as complete
        db_query("UPDATE {mover_map_references} SET complete=1 WHERE map_id=%d", $map_id);
      }
      break;
    case 'user':
      // Process reference fields on users? This is D7+ only.
      if (is_array($references['user'])) {
        // Nothing to do on Drupal 6
        drupal_set_message(t('Cannot save reference field for user @uid on Drupal 6.', array('@uid' => $id)));
      }
      break;
  }
}
